package game

import (
	"fmt"
	"net"
)

const (
	// How many packets we will buffer in case our consumer can't keep up.
	ChannelBufferSize = 4096

	// The maximum data length of a UDP datagram (8 bytes UDP header, 20 bytes IP header).
	UDPDataMaxLength = 65535 - 8 - 20
)

// Server represents a UDP server.
type Server struct {
	handlers    [256]Handler
	socket      *net.UDPConn
	incoming    chan Packet
	packetCount int
}

// NewServer creates a new server.
func NewServer() *Server {
	return &Server{
		incoming: make(chan Packet, ChannelBufferSize),
	}
}

// SetHandler sets the handler for the given byte code.
func (s *Server) SetHandler(code byte, handler Handler) {
	s.handlers[code] = handler
}

// Run starts the server on the given port.
func (s *Server) Run(port int) {
	s.socket = listen(port)
	defer s.socket.Close()

	s.read()
}

// PacketCount returns the number of processed packets.
func (s *Server) PacketCount() int {
	return s.packetCount
}

// ResetPacketCount resets the number of processed packets to zero.
func (s *Server) ResetPacketCount() {
	s.packetCount = 0
}

// Send sends the data prefixed with the byte code to the client.
func (s *Server) Send(code byte, data []byte, address *net.UDPAddr) error {
	tmp := make([]byte, len(data)+1)
	tmp[0] = code
	copy(tmp[1:], data)
	_, err := s.socket.WriteToUDP(tmp, address)
	return err
}

// listen creates a socket for the server and starts listening for incoming packets.
func listen(port int) *net.UDPConn {
	addr, err := net.ResolveUDPAddr("udp", fmt.Sprintf(":%d", port))

	if err != nil {
		panic(err)
	}

	connection, err := net.ListenUDP("udp", addr)

	if err != nil {
		panic(err)
	}

	return connection
}

// read is a blocking call which will read incoming packets and send them to a channel.
func (s *Server) read() {
	buffer := make([]byte, UDPDataMaxLength)

	for {
		n, addr, err := s.socket.ReadFromUDP(buffer)

		if err != nil {
			fmt.Println("Error reading from UDP connection:", err)
			continue
		}

		if n == 0 {
			continue
		}

		tmp := make([]byte, n)
		copy(tmp, buffer)
		s.incoming <- Packet{Data: tmp, Address: addr}
		s.packetCount++
	}
}
